from functools import partial
from typing import Tuple, List, Callable, Optional

import numpy as onp
import jax
import jax.numpy as np
from jaxtyping import Array, Float, PyTree, Int

from utils import *


XArr = Float[Array, "batch_size n_inp"]
YArr = Float[Array, "batch_size"]  # class label for classification
DTup = Tuple[XArr, YArr]
Params = Tuple[Float[Array, "n_out n_inp"], Float[Array, "n_out"]]  # (W, b)


def gen_classif_data(
    n_training_domains, dim_inv, dim_spu, n_train, sigma_x=0.6, mu_a=1., mu_b=0., spu_mixing=0.,
    data_rng=None, dist_rng=None
) -> Tuple[Tuple[List[DTup], Callable[[int], DTup]], Params, Params]:
    """
    return jnp.ndarray's
    """
    
    data_rng = data_rng or onp.random.RandomState(23)
    dist_rng = dist_rng or onp.random.RandomState(23)
    mu_inv = dist_rng.normal(size=(dim_inv,)) * 2
    X_sd = onp.concatenate([
        onp.ones((dim_inv,)) * 2.5, onp.ones((dim_spu,))]) * sigma_x
    
    def gen_domain(n, mu_spu=None):
        if mu_spu is None:
            mu_spu = dist_rng.normal(size=(dim_spu,))
        mu = onp.concatenate([mu_inv, mu_spu])
        # Y = data_rng.randint(0, 2, (n,))
        Y = onp.concatenate([np.zeros((n//2,)), np.ones((n//2,))]).astype('i')
        data_rng.shuffle(Y)
        X_mean = (Y*2-1)[:, None] * mu[None, :]
        x_eps = data_rng.normal(size=X_mean.shape)
        if spu_mixing > 0:
            A = dist_rng.normal(size=(dim_inv, dim_spu)) / dim_inv**0.5
            a, b = (1-spu_mixing**2)**0.5, spu_mixing
            x_eps[:, dim_inv:] = x_eps[:, dim_inv:] * a + x_eps[:, :dim_inv] @ A * b
        X = x_eps * X_sd + X_mean
        return jax.tree_map(jax.numpy.asarray, ((X, Y), mu_spu))
    
    train_dtups, train_mus = [], []
    for _ in range(n_training_domains):
        dt, mu = gen_domain(n_train)
        train_dtups.append(dt)
        train_mus.append(mu)
        
    mu_spu_test = -sum(train_mus) / n_training_domains * 0.5 * mu_a +\
        dist_rng.normal(size=(dim_spu,)) * mu_b
    dtups = (train_dtups, lambda n: gen_domain(n, mu_spu_test)[0])
           
    def to_param(mean):
        W = np.asarray(
            onp.concatenate([0*mean[:, None], mean[:, None]], 1) * 2 / X_sd[:, None]**2)
        b = np.zeros((2,))
        return W.T, b

    mean_inv = onp.concatenate([mu_inv, 0*mu_spu_test])
    mean_spu = onp.concatenate([mu_inv, mu_spu_test])
    
    return dtups, to_param(mean_inv), to_param(mean_spu)


def gen_regression_data(
    n_training_domains, dim_inv, dim_spu, n_train,
    inv_strength=2, noise_sd=0.5, mu_a=1., mu_b=0., spu_mixing=0.,
    data_rng=None, dist_rng=None
) -> Tuple[Tuple[List[DTup], Callable[[int], DTup]], Params, Params, float]:
    """
    return jnp.ndarray's
    """
    data_rng = data_rng or onp.random.RandomState(23)
    dist_rng = dist_rng or onp.random.RandomState(23)

    mu_inv = dist_rng.normal(size=(dim_inv,)) * inv_strength / dim_inv**0.5
    # note the scaling must be consistent across train/test envs
    y_var = inv_strength**2 + 1 + noise_sd**2
    
    def gen_domain(n, mu_spu=None):
        if mu_spu is None:
            mu_spu = dist_rng.normal(size=(dim_spu,)) / dim_spu**0.5
        mu = onp.concatenate([mu_inv, mu_spu])
        X = data_rng.normal(size=(n, dim_inv+dim_spu))
        if spu_mixing > 0:
            A = dist_rng.normal(size=(dim_inv, dim_spu)) / dim_inv**0.5
            a, b = (1-spu_mixing**2)**0.5, spu_mixing
            X[:, dim_inv:] = X[:, dim_inv:] * a + X[:, :dim_inv] @ A * b
        Y = (X @ mu[:, None]).squeeze(-1) + data_rng.normal(size=(n,)) * noise_sd
        Y = Y / y_var**0.5
        return jax.tree_map(jax.numpy.asarray, ((X, Y), mu_spu))
    
    train_dtups, train_mus = [], []
    for _ in range(n_training_domains):
        dt, mu = gen_domain(n_train)
        train_dtups.append(dt)
        train_mus.append(mu)
        
    mu_spu_test = -sum(train_mus) / n_training_domains * 0.5 * mu_a +\
        dist_rng.normal(size=(dim_spu,), scale=1/dim_spu**0.5) * mu_b
    dtups = (train_dtups, lambda n: gen_domain(n, mu_spu_test)[0])
           
    def to_param(mean):
        return np.asarray(mean[None] / y_var**0.5), np.zeros((1,))

    mean_inv = onp.concatenate([mu_inv, 0*mu_spu_test])
    mean_spu = onp.concatenate([mu_inv, mu_spu_test])
    
    return dtups, to_param(mean_inv), to_param(mean_spu), noise_sd**2/y_var