import os, sys, pickle
from functools import partial
from typing import Tuple, List, Callable, Optional
from dataclasses import dataclass, field

import numpy as onp
import jax
import jax.numpy as np
from jaxtyping import Array, Float, PyTree, Int
from tqdm import trange
from experiments.utils import s_parser, preflight

from utils import *
from data import gen_classif_data, gen_regression_data


Params = Tuple[Float[Array, "n_out n_inp"], Float[Array, "n_out"]]  # a single set of linear layer parameters
ChainParams = Tuple[Float[Array, "n_mc n_out n_inp"], Float[Array, "n_mc n_out"]]  # `n_mc` sets of parameters
Constraint = Callable[[Params], bool]
XArr = Float[Array, "batch_size n_inp"]  # input tensor
YArr = Float[Array, "batch_size"]  # output. represents class label for classification
DTup = Tuple[XArr, YArr]


def norm2(t: PyTree[Float]) -> Float:
    return sum(jax.tree_map(lambda a: (a**2).sum(), t))


# classification-specific functions

def _lr_predict(params: Params, x: XArr) -> Float[Array, "batch_size n_out"]:
    w, b = params
    logits = x @ w.T + b
    return logits - np.expand_dims(jax.scipy.special.logsumexp(logits, axis=-1), -1)

def classif_error(params: Params, data: DTup) -> Float:
    xs, ys = data
    p = _lr_predict(params, xs)
    correct = (p.argmax(-1) == ys).sum()
    total = p.shape[0]
    return 1 - correct / total

def chain_accuracy(cparams: ChainParams, data: DTup) -> Float:
    xs, ys = data
    p = np.exp(jax.vmap(_lr_predict, in_axes=(0, None))(cparams, xs)).mean(0)
    correct = (p.argmax(-1) == ys).sum()
    total = p.shape[0]
    return correct / total

def avg_ce_loss(params: Params, dat: DTup) -> Float:
    (x, y) = dat
    logprobs = _lr_predict(params, x)
    nll = np.take_along_axis(logprobs, np.expand_dims(y, axis=1), axis=1)
    return -np.mean(nll)


# regression-specific functions

def _linear_predict(params: Params, x: XArr):
    return x @ params[0].T + params[1]

def pred_mse(params: Params, dat: DTup) -> Float:
    return mse(_linear_predict(params, dat[0]), dat[1][:, None])

def chain_pred_mse(cparams: ChainParams, dat: DTup) -> Float:
    (Wc, bc), (x, y) = cparams, dat
    chain_pred = np.einsum("moi,bi->mbo", Wc, x) + bc[:, None]
    return mse(chain_pred.mean(0), y[:, None])

def gen_err_constraint(train_dtups: List[DTup],
                       err_fn: Callable[[Params, DTup], Float],
                       threshold: float) -> Constraint:
    def ret(param: Params) -> bool:
        errs = []
        for dtup in train_dtups:
            errs.append(err_fn(param, dtup))
        return np.max(np.asarray(errs)) <= threshold
    return ret


# The sampler

def _mala_transition_log_prob(params: Params, params1: Params, grads: Params, eta: Float):
    # params1 | params ~ N(params + eta grads, 2*eta I)
    return jax.tree_util.tree_reduce(
        lambda a,b: a+b,
        jax.tree_map(
            lambda p0, p1, g: -((p1-p0-eta*g)**2).sum() / (2*2*eta),
            params, params1, grads),
        0.)


@partial(jax.jit, static_argnames=('log_lh_fn', 'constraint_fn'))
def _c_lmc_step_fn(params: Params,
                   dtup0: DTup,  # first arg to log_lh_fn
                   dtup1: DTup,  # second arg
                   key: jax.random.PRNGKeyArray,
                   stepsize: float,
                   temp: float,
                   log_lh_fn: Callable[[Params, DTup, DTup], Float],
                   constraint_fn: Optional[Constraint]  # None -- unconstrained MALA
                  ) -> Tuple[Params, Float]:

    get_lh_and_grad = jax.value_and_grad(log_lh_fn)
    log_joint, grads = get_lh_and_grad(params, dtup0, dtup1)

    leaves, treedef = jax.tree_util.tree_flatten(params)
    key1, key2 = jax.random.split(key)
    keys = jax.random.split(key1, len(leaves))
    noises = [jax.random.normal(ckey, shape=leaf.shape) for leaf, ckey in zip(leaves, keys)]
    noises = jax.tree_util.tree_unflatten(treedef, noises)

    def ld_step(eta):
        return jax.tree_map(lambda p, g, n: p + eta*g + temp*np.sqrt(2*eta)*n, params, grads, noises)
    
    if constraint_fn is None:  # do MALA then
        params1 = ld_step(stepsize)
        log_joint_1, grads1 = get_lh_and_grad(params1, dtup0, dtup1)
        mala_logprob = (
            log_joint_1 + _mala_transition_log_prob(params1, params, grads1, stepsize)) - (
            log_joint + _mala_transition_log_prob(params, params1, grads, stepsize))
        return jax.lax.cond(
            jax.random.uniform(key2) < np.exp(np.minimum(0., mala_logprob)),
            lambda: (params1, stepsize),
            lambda: (params, 0.)
        )

    THRESHOLD = stepsize / (2**8)
    def loop_cond(eta):
        cur_param = ld_step(eta)
        return np.logical_and(eta > THRESHOLD, np.logical_not(constraint_fn(cur_param)))

    eta = jax.lax.while_loop(loop_cond, lambda e: e / 2., stepsize)
    return jax.lax.cond(eta > THRESHOLD, lambda: ld_step(eta), lambda: params), eta


def _do_LMC(paramss: ChainParams,
                adapt_dtup: DTup,
                train_dtup: DTup,
                pkey: jax.random.PRNGKeyArray,
                n_iters: int,
                stepsize: float,
                temperature: float,
                loglh_fn: Callable[[Params, DTup, DTup], Float],
                constraint_fn=None) -> ChainParams:
    
    n_chains = jax.tree_util.tree_flatten(paramss)[0][0].shape[0]
    step_fn = partial(
        _c_lmc_step_fn, stepsize=stepsize, temp=temperature,
        constraint_fn=constraint_fn, log_lh_fn=loglh_fn)
    displacements = []
    
    for _ in trange(n_iters):
        pkey, ckey = jax.random.split(pkey)
        ckeys = jax.random.split(ckey, n_chains)
        paramss, etas = jax.vmap(
            step_fn, in_axes=(0, None, None, 0))(paramss, adapt_dtup, train_dtup, ckeys)
        displacements.append(onp.asarray(np.mean(etas)))
        
    return paramss, onp.mean(displacements)


def concat_dtups(dtups):
    CC = partial(onp.concatenate, axis=0)
    return (
        CC([d[0] for d in dtups]),
        CC([d[1] for d in dtups]))

# classification initialization (Logistic regression)

def lr_init(dtup: DTup, weights=None) -> Params:
    from sklearn.linear_model import LogisticRegression
    lrmodel = LogisticRegression(max_iter=10000)
    lrmodel.fit(*dtup, sample_weight=weights)
    coef = lrmodel.coef_ 
    assert coef.shape[0] == 1
    return onp.concatenate([0*coef, coef], 0), np.zeros((2,))


def gen_blr_init(Dtr_weight):

    def init_fn(dtups) -> Params:
        if Dtr_weight == 0:
            return lr_init(dtups[0])
        elif Dtr_weight == 'infty' or Dtr_weight == 'full':
            return lr_init(dtups[1])
        ((Xa, Ya), (Xtr, Ytr)) = dtups
        dtup = (onp.concatenate([Xa, Xtr], 0), onp.concatenate([Ya, Ytr], 0))
        ratio = Xa.shape[0] / Xtr.shape[0] * Dtr_weight
        weights = onp.concatenate([onp.ones((Xa.shape[0],)), onp.ones((Xtr.shape[0],))*ratio])
        return lr_init(dtup, weights)

    return init_fn

# regression initialization (OLS)

def gen_ols_init(Dtr_weight):

    def ols_init(X, Y, weights=None) -> Params:
        from sklearn.linear_model import RidgeCV
        M = RidgeCV(alphas=onp.logspace(0.1, 1e2, 6))
        M.fit(X, Y, sample_weight=weights)
        w = M.coef_[None]
        return (np.asarray(w), np.zeros((1,)))

    def init_fn(dtups) -> Params:
        if Dtr_weight == 0:
            return ols_init(*dtups[0])
        if Dtr_weight == 'infty' or Dtr_weight == 'full':
            return ols_init(*dtups[1])
        ((Xa, Ya), (Xtr, Ytr)) = dtups
        dtup = (onp.concatenate([Xa, Xtr], 0), onp.concatenate([Ya, Ytr], 0))
        ratio = Xa.shape[0] / Xtr.shape[0] * Dtr_weight
        weights = onp.concatenate([onp.ones((Xa.shape[0],)), onp.ones((Xtr.shape[0],))*ratio])
        return ols_init(*dtup, weights)
    
    return init_fn

def tile_params(pa: Params, n_mc: int) -> ChainParams:
    return jax.tree_map(lambda p: p[None].repeat(n_mc, 0), pa)


@dataclass
class ExpConfig:
    n_iters: int = 2500
    step_size_base: float = 0.001
    n_mc_chains: int = 50
    prior_sd_scale: float = 4.
    # 
    constraint_errs: List[float] = field(default_factory=[0.05, 0.1, 0.2].copy)
    modified_lh_ratio: List[float] = field(default_factory=[0.3, 1].copy)
    modified_prior_strength: List[float] = field(default_factory=[0.3, 1, 3, 10, 30].copy)
    reinit_threshold: int = 3
    main_seed: int = 23
    # 
    mode: str = 'classification'
    dist_seed: int = 23
    n_adapt_max: int = 256
    n_adapt_min: int = 2
    grid_size: int = 8
    n_train_domains: int = 3
    n_train: int = 3000
    dim_inv: int = 20
    dim_spu: int = 50
    x_sd_scale: float = 3.  # for classification
    inv_strength: float = 2.  # for regression
    regr_noise_sd: float = 0.5  # for regression
    spu_mixing: float = 0.
    mu_a: float = 1.
    mu_b: float = 0.


def gen_mse_fn(train_dtups: List[DTup]) -> Callable[[Params], Float[Array, "n_envs"]]:

    def gen_suff_stats(Xe: XArr, Ye: YArr):
        Xe = np.concatenate([Xe, np.ones_like(Xe[:, :1])], 1)
        Ye = Ye[:, None]  # [n, 1]
        return Xe.T @ Xe, Xe.T @ Ye, (Ye**2).sum(), Xe.shape[0]  # (d+1, d+1), (d+1, 1), ()

    suff_stats = [gen_suff_stats(*dtup) for dtup in train_dtups]

    def ret(params: Params):
        W, b = params  # (1, d), (1,)
        theta = np.concatenate([W, b[:, None]], 1).T  # (d+1, 1)
        mses = []
        for S, B, ys, n_e in suff_stats:
            mse = ((theta.T @ S @ theta).sum() - 2 * (theta * B).sum() + ys) / n_e
            mses.append(mse)
        return np.asarray(mses)

    return ret


def main(args: Any):

    cfg: ExpConfig = args.c

    o_rng = onp.random.RandomState(cfg.main_seed)

    if cfg.mode == 'regression':
        (Dtrs, gen_Dte_fn), true_param_inv, true_param_spu, noise_var = gen_regression_data(
            cfg.n_train_domains, cfg.dim_inv, cfg.dim_spu, cfg.n_train,
            cfg.inv_strength, cfg.regr_noise_sd, cfg.mu_a, cfg.mu_b, cfg.spu_mixing,
            data_rng=o_rng, dist_rng=onp.random.RandomState(cfg.dist_seed))
        test_fn, chain_test_fn = pred_mse, chain_pred_mse
        avg_nll_fn = lambda param, dtup: pred_mse(param, dtup) / (2*noise_var)
        gen_init_fn = gen_ols_init
        base_test_err_diff = 0.05
    elif cfg.mode == 'classification':
        (Dtrs, gen_Dte_fn), true_param_inv, true_param_spu = gen_classif_data(
            cfg.n_train_domains, cfg.dim_inv, cfg.dim_spu, cfg.n_train, cfg.x_sd_scale,
            cfg.mu_a, cfg.mu_b, data_rng=o_rng,
            dist_rng=onp.random.RandomState(cfg.dist_seed))
        test_fn, avg_nll_fn, chain_test_fn = classif_error, avg_ce_loss, chain_accuracy
        gen_init_fn = gen_blr_init
        base_test_err_diff = 0.02
    else:
        raise ValueError(cfg.mode)

    if cfg.mode == 'classification':
        gen_constraint = partial(gen_err_constraint, train_dtups=Dtrs, err_fn=test_fn)
    else:
        train_mse_fn = gen_mse_fn(Dtrs)
        def gen_constraint(threshold):
            return (lambda params: np.max(train_mse_fn(params)) < threshold)

        def train_avg_nll(params):
            mses = train_mse_fn(params)
            n_es = np.asarray([dtr[0].shape[0] for dtr in Dtrs])
            return (mses*n_es).sum() / (2*noise_var*n_es.sum())

    Dte = gen_Dte_fn(10000)
    Dtr_concat = concat_dtups(Dtrs)

    true_worst_case_err = max(test_fn(true_param_inv, dtup) for dtup in Dtrs)
    print('train err (true inv):', test_fn(true_param_inv, Dtr_concat), true_worst_case_err)
    print('OD test loss/err (true inv):',
          avg_nll_fn(true_param_inv, Dte), test_fn(true_param_inv, Dte))
    print('OD test loss/err (true spu):',
          avg_nll_fn(true_param_spu, Dte), test_fn(true_param_spu, Dte))

    initial_params = gen_init_fn(Dtr_weight='infty')((None, Dtr_concat))
    print('train loss/err (ERM init):',
          avg_nll_fn(initial_params, Dtr_concat), test_fn(initial_params, Dtr_concat))
    print('OD test loss/err (ERM init):',
          avg_nll_fn(initial_params, Dte), test_fn(initial_params, Dte))

    W_PRIOR_VAR = (initial_params[0]**2).mean() * cfg.prior_sd_scale

    # specify tasks
    n_tot_train = Dtr_concat[0].shape[0]

    def interp_log_joint(param, dadapt, dtrain, lh_ratio):
        n_adapt = dadapt[0].shape[0]
        loglh = -avg_nll_fn(param, dadapt) * n_adapt
        if lh_ratio == 'full' or lh_ratio > 0:
            if not cfg.mode.startswith('regr'):
                avg_llh2 = -avg_nll_fn(param, dtrain)
            else:
                avg_llh2 = -train_avg_nll(param)
            scale = n_adapt * lh_ratio if lh_ratio != 'full' else n_tot_train
            loglh = loglh + avg_llh2 * scale
        return loglh - norm2(param) / (2*W_PRIOR_VAR)

    def mprior_log_joint(param, dadapt, _, erm_param, prior_strength):
        loglh = -avg_nll_fn(param, dadapt) * n_adapt
        prior_var = W_PRIOR_VAR / prior_strength
        log_prior = -norm2(jax.tree_map(lambda a,b: a-b, param, erm_param)) / (2*prior_var)
        return loglh + log_prior

    std_loglh_fn = partial(interp_log_joint, lh_ratio=0)

    tasks = []

    # CBLR
    for err_tol in [base_test_err_diff] + cfg.constraint_errs:
        c_fn = gen_constraint(threshold=true_worst_case_err+err_tol)
        # special case with n_adapt == 0
        # if err_tol == base_test_err_diff:
        #     tasks.append((
        #         f'crw',
        #         partial(_do_LMC, temperature=1., loglh_fn=lambda *_: 0., constraint_fn=c_fn),
        #         lambda _: initial_params
        #     ))
        tasks.append((
            f'cblr-{err_tol}',
            partial(_do_LMC, temperature=1., loglh_fn=std_loglh_fn, constraint_fn=c_fn),
            lambda _: initial_params
        ))

    # BLR and BLR-LC
    for interp_ratio in [0, 'full'] + cfg.modified_lh_ratio:
        l_fn = partial(interp_log_joint, lh_ratio=interp_ratio)
        tasks.append((
            f'blr-{interp_ratio}',
            partial(_do_LMC, temperature=1., constraint_fn=None, loglh_fn=l_fn),
            gen_init_fn(Dtr_weight=interp_ratio)
        ))

    # BLR-Prior-
    for pstr in cfg.modified_prior_strength:
        l_fn = partial(mprior_log_joint, erm_param=initial_params, prior_strength=pstr)
        tasks.append((
            f'blr-prior-{pstr}',
            partial(_do_LMC, temperature=1., constraint_fn=None, loglh_fn=l_fn),
            lambda _: initial_params
        ))

    # generate adaption data. for classification, also make sure it is class-balanced
    Da_all = gen_Dte_fn(cfg.n_adapt_max)
    n_adapts = (log_linspace(cfg.n_adapt_min, cfg.n_adapt_max, cfg.grid_size).astype('i') + 1
                ) // 2 * 2
    print(n_adapts)
    if cfg.mode == 'classification':
        Ya = Da_all[1]
        idcss = (onp.where(Ya==0)[0], onp.where(Ya==1)[0])
        idcs = onp.concatenate([idcss[0][:, None], idcss[1][:, None]], 1).reshape((-1,))
        for _ in [2, 4, 8]:
            assert onp.abs(Ya[idcs][:_].mean() - 0.5) < 1e-3
        Da_all = (Da_all[0][idcs], Da_all[1][idcs])

    # run all tasks
    chain_snapshot = {name: None for name, _, __ in tasks}
    to_dump = {name: [] for name, _, __ in tasks}
    pkey = jax.random.PRNGKey(cfg.main_seed)
    for n_adapt in n_adapts:
        Dadapt = (Da_all[0][:n_adapt], Da_all[1][:n_adapt])
        pkey, ckey = jax.random.split(pkey)
        for name, update_fn, init_fn in tasks:
            n_iters = cfg.n_iters*4 if name.startswith('c') else cfg.n_iters
            if n_adapt < cfg.reinit_threshold:
                paramss = tile_params(init_fn((Dadapt, Dtr_concat)), cfg.n_mc_chains)
            else:
                paramss = chain_snapshot[name]
                n_iters = n_iters // 3
            init_acc = chain_test_fn(paramss, Dte)
            ssz = cfg.step_size_base * min(1, 16 / n_adapt)
            if name == 'blr-full':
                ssz = min(ssz, cfg.step_size_base / n_tot_train)
            nparamss, displacement = update_fn(paramss, Dadapt, Dtr_concat, ckey, n_iters, ssz)
            chain_snapshot[name] = nparamss
            acc = chain_test_fn(nparamss, Dte)
            print(n_adapt, name, init_acc, acc, displacement, norm2(paramss), norm2(nparamss))
            to_dump[name].append((acc, init_acc, displacement))

    with open(os.path.join(args.dir, "res.pkl"), "wb") as fout:
        pickle.dump(to_py_dict(to_dump), fout)


if __name__ == '__main__':
    parser = s_parser("dgsim")
    parser.add_arguments(ExpConfig, dest='c')
    args = parser.parse_args()
    preflight(args)
    main(args)
