import numpy as np, os, sys, pickle, json, matplotlib, argparse
from matplotlib import pyplot as plt

matplotlib.rcParams.update({'font.size': 13})
matplotlib.use('Agg')

parser = argparse.ArgumentParser()
parser.add_argument('--classif_path', type=str, default='')
parser.add_argument('--regr_path', type=str, default='')
parser.add_argument('--dist_seed', type=int, default=1)
parser.add_argument('--fig_number', type=str, default='1', choices=['1', 'S1', 'S2'])
args = parser.parse_args()

def compatible(hps, crit):
    if not isinstance(hps, dict):
        if callable(crit):
            return crit(hps)
        if isinstance(crit, float):
            return np.abs(hps - crit) < 1e-7
        return hps == crit
    for k in crit:
        if k not in hps:
            print(k, hps)
            return False
        if not compatible(hps[k], crit[k]):
            return False
    return True

def load(base_path, criterion):
    ret = []
    for path in os.listdir(base_path):
        path = os.path.join(base_path, path)
        if not os.path.exists(os.path.join(path, 'res.pkl')):
            continue
        with open(os.path.join(path, 'hps.txt')) as fin:
            hps = json.loads(fin.readline())
        if not compatible(hps, criterion):
            #print(hps)
            continue
        
        # get N and baselines
        with open(os.path.join(path, 'stdout')) as fin:
            lines = fin.readlines()
        for ln in lines:
            if ln.startswith('['):
                ln = ln.strip()[1:-1].split(' ')
                Ns = [int(a) for a in ln if a != ""]
        
        with open(os.path.join(path, 'res.pkl'), 'rb') as fin:
            dump = pickle.load(fin)
            
        for label, offset in [('_inv', 2), ('_spu', 3), ('_erm', 5)]:
            err = float(lines[offset].split(': ')[1].split(' ')[1])
            dump[label] = [(1-err, -1, -1) for _ in Ns]

        ret.append(dump)
    return ret, Ns

def aggregate_dumps(dumps):
    trace_ag = {}

    for dump in dumps:
        for method, trace in dump.items():
            acc_trace = [it[0] for it in trace]
            if method not in trace_ag:
                trace_ag[method] = [[a] for a in acc_trace]
            else:
                for i, a in enumerate(acc_trace):
                    trace_ag[method][i].append(a)

    return trace_ag


def plot_dump(dumps, typ='pct', legend_map=None, legend=True, do_label=True, is_acc=True, pct=20):
    
    trace_ag = {
        meth:v for meth, v in aggregate_dumps(dumps).items() if not(
            any(meth.startswith(k) for k in ['clr', '_']) 
            or (legend_map is not None and meth not in legend_map))}
    tot = len(trace_ag)
    
    for i, (meth, dat) in enumerate(trace_ag.items()):
        dat = 1 - np.asarray(dat) if is_acc else np.asarray(dat)
        if legend_map is None:
            label = meth
            ls, alf = {
                'blr': ('--', 0.025),
                'crw': (':', 0.05),
                'clr': (':', 0.05),
                'cblr': ('-', 0.3)
            }[meth.split('-')[0]]
            if meth.startswith('blr-p'):
                ls = ':'
        else:
            label, ls, alf = legend_map[meth]
        if typ == 'pct':
            ct = np.median(dat, axis=1)
            lo = np.percentile(dat, pct, axis=1)
            hi = np.percentile(dat, 100-pct, axis=1)
        else:
            ct = dat.mean(1)
            lo = dat.mean(1) - dat.std(1)
            hi = dat.mean(1) + dat.std(1)
        # plt.plot(Ns, 1-ct, label=label, linestyle=ls, marker='+')
        #plt.fill_between(Ns, 1-hi, 1-lo, alpha=alf)
        xs = np.asarray(Ns) * np.exp((i/tot-0.5)*0.1)
        capsize = 3 if meth.startswith('cblr') else None
        if not do_label:
            label = None
        eb = plt.errorbar(
            xs, ct, yerr=(ct-lo, hi-ct), label=label, linestyle=ls, capsize=capsize, linewidth=1.5+alf)
        eb[-1][0].set_linestyle(ls)
    
    if legend:
        plt.legend()
    #plt.xticks(Ns)
    plt.xscale('log')
    #plt.yscale('log')
    plt.xticks(Ns)
    plt.gca().get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
    
    sys.stderr.write(f'{len(dumps)} replications plotted\n')


# Fig 1 / S1
crit_6k = {
    'step_size_base': 1e-3,
    'mu_a': 1, 'x_sd_scale': 3., 'dist_seed': args.dist_seed,
    'n_train': 6000
}
crit_classif_large = {
    'mu_a': 1, 'x_sd_scale': 3., 'dist_seed': args.dist_seed,
    'n_train': 12000
}
crit_regr_large = {
    'mu_a': 1, 'x_sd_scale': 3., 'dist_seed': args.dist_seed,
    'n_train': 18000, 'spu_mixing': 0.
}

if args.fig_number == '1':
    legend_map = {
        'blr-0': ('BLR (adapt. only)', (5,(7,3)), 0.5), 
        'blr-1': ('BLR-LC-N_adapt', '--', 0.5),
        'blr-full': ('BLR-LC-N_train', ':', 0.5),
        'blr-prior-3': ('BLR-Prior-3', '-.', 0.5),
        'cblr-0.05': ('CBLR-0.05', '-', 1.5),
        'cblr-0.1': ('CBLR-0.1', '-', 1.5),
        'cblr-0.4': ('CBLR-0.4', (0,(3,1,1,1,1)), 0.5)
    }
else:
    legend_map = {
        'blr-0': ('BLR (adapt. only)', (5,(7,3)), 0.5), 
        'blr-0.3': ('BLR-LC-0.3_Nadapt', ':', 0.5),
        'blr-1': ('BLR-LC-Nadapt', '--', 0.5),
        'blr-full': ('BLR-LC-N_train', ':', 0.5),
        'blr-prior-1': ('BLR-Prior-1', '-.', 0.5),
        'blr-prior-3': ('BLR-Prior-3', '-.', 0.5),
        'blr-prior-10': ('BLR-Prior-10', '-.', 0.5),
        'cblr-0.05': ('CBLR-0.05', '-', 1.5),
        'cblr-0.1': ('CBLR-0.1', '-', 1.5),
        'cblr-0.2': ('CBLR-0.2', (0,(3,1,1)), 0.5),
        'cblr-0.4': ('CBLR-0.4', (0,(3,1,1,1,1)), 0.5)
    }

baseline_legend = {
    "_erm": 'ERM (train. only)',
    # "_spu": '(Bayes err)',
    #"_inv": '(true inv. predictor)',
    #'crw': 'CRW (train. only)'
}

def TWOS(a): return np.asarray([a,a])

fig = plt.figure(figsize=(12, 6), facecolor='w')

has_classif = os.path.exists(args.classif_path)

if has_classif:
    plt.subplot(121)
    dumps, Ns = load(
        args.classif_path, {
            'c': crit_6k if args.fig_number in ['1', 'S1'] else crit_classif_large
        })
    trace_ag = aggregate_dumps(dumps)
    plot_dump(dumps, legend_map=legend_map, legend=False)
    
    for k, legend in baseline_legend.items():
        errs = 1 - np.asarray(trace_ag[k][0])
        lo, med, hi = np.percentile(errs, [20, 50, 80])
        xs = np.asarray([Ns[0]*0.9, Ns[-1]*1.1])
        plt.fill_between(xs, TWOS(lo), TWOS(hi), alpha=0.4, label=legend, color='gray')
    
    plt.xlabel('N_adapt')
    plt.ylabel('classification error')
    plt.yscale('log')
    plt.title('Classification')
    fmt = matplotlib.ticker.ScalarFormatter()
    fmt.set_scientific(False)
    plt.gca().yaxis.set_major_formatter(fmt)
    plt.gca().yaxis.set_minor_formatter(fmt)
    plt.xlim(*xs)
else:
    print('skipping invalid classification path')

if os.path.exists(args.regr_path):
    if has_classif:
        plt.subplot(122)
    dumps, Ns = load(
        args.regr_path, {
            'c': crit_6k if args.fig_number in ['1', 'S1'] else crit_regr_large
        })
    trace_ag = aggregate_dumps(dumps)
    plot_dump(dumps, legend_map=legend_map, legend=False, is_acc=False,
              do_label=not has_classif)
    
    for k, legend in baseline_legend.items():
        errs = 1-np.asarray(trace_ag[k][0])
        lo, med, hi = np.percentile(errs, [20, 50, 80])
        xs = np.asarray([Ns[0]*0.9, Ns[-1]*1.1])
        plt.plot(xs, TWOS(med), linestyle='-', alpha=0.6, linewidth=3, label=None, color='gray')
    
    plt.xlabel('N_adapt')
    plt.ylabel('test MSE')
    plt.yscale('log')
    fmt = matplotlib.ticker.ScalarFormatter()
    fmt.set_scientific(False)
    ax = plt.gca()
    ax.yaxis.set_major_formatter(fmt)
    ax.yaxis.set_minor_formatter(fmt)
    plt.xlim(*xs)
    plt.title('Regression')

lgd = fig.legend(loc=7, ncol=1, bbox_to_anchor=(1.2, 0.55))
plt.tight_layout()
fig.savefig(f'/tmp/synth_fig{args.fig_number}.pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')


